settings = {
    'BOT_NAME': 'indeed',
    'SPIDER_MODULES': ['indeed.spiders'],
    'NEWSPIDER_MODULE': 'indeed.spiders',
    
    # Log level (optional if you want to reduce logging verbosity)
    'LOG_LEVEL': 'WARNING',  
    
    # ScrapeOps configuration
    'SCRAPEOPS_API_KEY': 'get_your_own_key',  # Replace with your actual ScrapeOps API Key
    'SCRAPEOPS_PROXY_ENABLED': True,

    # Obey robots.txt rules
    'ROBOTSTXT_OBEY': False,

    # Item pipeline
    'ITEM_PIPELINES': {
        "indeed.pipelines.RawDataPipeline": 200,
    },

    # Downloader middlewares
    'DOWNLOADER_MIDDLEWARES': {
        # ScrapeOps Retry Middleware
        'scrapeops_scrapy.middleware.retry.RetryMiddleware': 550,
        'scrapy.downloadermiddlewares.retry.RetryMiddleware': None,

        # ScrapeOps Proxy Middleware
        'indeed.middlewares.ScrapeOpsProxyMiddleware': 725,
    },

    # ScrapeOps monitoring extension
    'EXTENSIONS': {
        'scrapeops_scrapy.extension.ScrapeOpsMonitor': 500,
    },

    # Concurrency settings (adapt based on your ScrapeOps plan)
    'CONCURRENT_REQUESTS': 1,  # Max concurrency on ScrapeOps Free Plan is 1 thread
    
    # Additional configurations for Scrapy (Optional - similar to your reference format)
    'REQUEST_FINGERPRINTER_IMPLEMENTATION': "2.7",
    'TWISTED_REACTOR': "twisted.internet.asyncioreactor.AsyncioSelectorReactor",
    'FEED_EXPORT_ENCODING': "utf-8",
}



# Define your item pipelines here
#
# Don't forget to add your pipeline to the ITEM_PIPELINES setting
# See: https://docs.scrapy.org/en/latest/topics/item-pipeline.html


# useful for handling different item types with a single interface
from itemadapter import ItemAdapter
import json
from datetime import datetime
from scrapy.signalmanager import dispatcher
from scrapy import signals

class MergedDataPipeline:
    def __init__(self):
        self.raw_data = []
        self.cleaned_data = []
    def process_item(self,item,spider):
        adapter = ItemAdapter(item)

        if adapter.get('source'):
            #Process raw data
            self.raw_data.append(adapter.asdict())

            #Process cleaned data
            cleaned_item = self.clean_item(item)
            self.cleaned_data.append(cleaned_item)
        return item
    def close_spider(self,spider):

        #Process data
        spider.log("pipeline Closing spider and sending custom_closed_signal...")
        dispatcher.send(signal=signals.spider_closed, 
                        sender=spider, 
                        cleaned_data = self.cleaned_data, 
                        raw_data = self.raw_data)
        
    def clean_item (self,item):

        if isinstance(item, dict):
            cleaned_item = {}
            for key, value in item.items():
                cleaned_value = self.clean_value(key, value)
                if cleaned_value is not None:
                    cleaned_item[key] = cleaned_value
            return cleaned_item
        elif isinstance(item,list):
            cleaned_list = []
            for element in item:
                cleaned_element = self.clean_item(element)
                if cleaned_element:
                    cleaned_list.append(cleaned_element)
            return cleaned_list
        else:
            return item
    
    def clean_value(self, key, value):

        float_keys = ['latitude','longitude']
        
        area_keys = ['property_living_area_sqm']

        date_keys = ['timeline_advertisement_creation']

        int_keys = ['property_construction_year','property_rooms_amount','property_floor_number',]

        special_str_keys = ['property_housing_type_finnish','address_full','source','address_postal_code',
                            'property_listing_type','address_country','property_sauna','property_heating',
                            'property_energy_label',
                            'property_additonal_terms','property_renovation_info','property_bathroom_features',
                            'property_kitchen_features','property_condition_finnish','property_configuration','property_description',
                            'property_rent_price_currency','property_sale_price_currency']

        if isinstance(value, int) or isinstance(value, float):

            if key in int_keys:

                try:

                    numeric_value = int(value)
                    
                    return numeric_value
                
                except:

                    return 0

            if key in area_keys:

                try:

                    cleaned_value = float(value)

                    return cleaned_value
                
                except:

                    pass
            else:
                return value

        if isinstance(value,str):

            if key in float_keys:

                try:

                    numeric_value = float(value)

                    return numeric_value
                
                except:

                    pass
        
            if key in int_keys:

                try:

                    numeric_value = int(value)

                    return numeric_value
                
                except:

                    pass
            
            elif key in date_keys:

                return self.convert_to_timestamp(value)
            
            elif key in special_str_keys:

                return value
            else:
                return value
            
        else:
            return value 
            
    def convert_to_timestamp(self,date_string):
        # Define the format of your date string
        date_format = '%Y-%m-%d %H:%M:%S'

        # Convert the date string to a datetime object
        date_object = datetime.strptime(date_string, date_format)

        # Convert the datetime object to a Unix timestamp
        timestamp = date_object.timestamp()

        return int(timestamp)
    
# class RawDataPipeline:

#     def __init__(self):
#         self.raw_data = []

#     def process_item(self, item, spider):
#         # Basic data validation: Check if the scraped item is not empty
#         adapter = ItemAdapter(item)
#         if adapter.get('source'):
#             self.raw_data.append(adapter.asdict())
#         return item

#     def close_spider(self, spider):
#         with open('asunnot_raw_data.json', 'w',encoding='utf-8') as file:
#             json.dump(self.raw_data, file, indent=2, ensure_ascii=False)

# class CleanedDataPipeline:

#     def __init__(self):

#         self.cleaned_data = []

#     def process_item(self, item, spider):

#         cleaned_item = self.clean_item(item)
#         self.cleaned_data.append(cleaned_item)
#         return item
    
#     def close_spider(self, spider):

#         with open('asunnot_cleaned_data.json', 'w' , encoding = 'utf-8') as file:
#             json.dump(self.cleaned_data, file, indent=2, ensure_ascii=False)
    
#     def clean_item (self,item):

#         if isinstance(item, dict):
#             cleaned_item = {}
#             for key, value in item.items():
#                 cleaned_value = self.clean_value(key, value)
#                 if cleaned_value is not None:
#                     cleaned_item[key] = cleaned_value
#             return cleaned_item
#         elif isinstance(item,list):
#             cleaned_list = []
#             for element in item:
#                 cleaned_element = self.clean_item(element)
#                 if cleaned_element:
#                     cleaned_list.append(cleaned_element)
#             return cleaned_list
#         else:
#             return item
    
#     def clean_value(self, key, value):

#         price_keys = ['price']

#         float_keys = ['latitude','longitude']
        
#         area_keys = ['size']

#         date_keys = ['creation_date']

#         int_keys = ['cardid','construction_hear','number_of_rooms']

#         special_str_keys = ['property_type','address','location_path','source','zipcode','or_rent','type_of_construction','address_country']

#         if isinstance(value, int) or isinstance(value, float):

#             if key in price_keys:
                
#                 try:

#                     numeric_value = float(value)

#                     cleaned_value ={
#                         f"{key}_value": numeric_value,
#                         f"{key}_currency": 'EUR',
#                     }

#                     return cleaned_value
                
#                 except:

#                     pass

#             if key in int_keys:

#                 try:

#                     numeric_value = int(value)
                    
#                     return numeric_value
                
#                 except:

#                     return 0

#             if key in area_keys:

#                 try:

#                     numeric_value = float(value)

#                     cleaned_value = {
#                         f"{key}_value": numeric_value,
#                         f"{key}_unit": 'square meter',
#                     }

#                     return cleaned_value
                
#                 except:

#                     pass

#         if isinstance(value,str):

#             if key in float_keys:

#                 try:

#                     numeric_value = float(value)

#                     return numeric_value
                
#                 except:

#                     pass
        
            
#             elif key in date_keys:

#                 return self.convert_to_timestamp(value)
            
#             elif key in special_str_keys:

#                 return value
            
#         else:
#             return value 
            
#     def convert_to_timestamp(self,date_string):
#         # Define the format of your date string
#         date_format = '%Y-%m-%d %H:%M:%S'

#         # Convert the date string to a datetime object
#         date_object = datetime.strptime(date_string, date_format)

#         # Convert the datetime object to a Unix timestamp
#         timestamp = date_object.timestamp()

#         return int(timestamp)


# Define your item pipelines here
#
# Don't forget to add your pipeline to the ITEM_PIPELINES setting
# See: https://docs.scrapy.org/en/latest/topics/item-pipeline.html


# useful for handling different item types with a single interface
from itemadapter import ItemAdapter
from datetime import datetime
import time
import json
import re
from scrapy.signalmanager import dispatcher
from scrapy import signals

def normalize_key (text):

    if text is None:
        return ''
    
    if text is not None:
        
        normalized_key = text.replace('(p.m.)','').replace('(s)','').strip()

        # Replace spaces with underscores
        normalized_key = normalized_key.replace(' ', '_').replace('.','').replace('-','_').strip()
    
        # Remove special characters
        normalized_key = ''.join(char for char in normalized_key if char.isalnum() or char in ['_', '-'])

        normalized_key = normalized_key.replace('__',"_")

        # Decapitalize the key
        normalized_key = normalized_key.lower()

        return normalized_key

class RawDataPipeline:

    def __init__(self):
        self.raw_data = []

    def process_item(self, item, spider):
        # Basic data validation: Check if the scraped item is not empty
        adapter = ItemAdapter(item)
        if adapter.get('source'):
            self.raw_data.append(adapter.asdict())
        return item

    def close_spider(self,spider):

        #Process data
        spider.log("pipeline Closing spider and sending custom_closed_signal...")
        dispatcher.send(signal=signals.spider_closed, 
                        sender=spider, 
                        raw_data = self.raw_data)



settings = {
    "BOT_NAME": "huurdatascraper",
    "SPIDER_MODULES": ["ClappScrapers.den_rental.spider"],
    "NEWSPIDER_MODULE": "ClappScrapers.den_rental.spider",
    "ROBOTSTXT_OBEY": False,
    "ITEM_PIPELINES": {
        "ClappScrapers.den_rental.spider.pipelines.MergedDataPipeline": 100,
    },
    'COOKIES_ENABLED' : False,
    'LOG_LEVEL':'WARNING',
    'CONCURRENT_REQUESTS' : 8,  # Reduce the number of concurrent requests
    'CONCURRENT_REQUESTS_PER_DOMAIN' : 4,  # Reduce the number of concurrent requests per domain
    'CONCURRENT_REQUESTS_PER_IP' : 4,  # Reduce the number of concurrent requests per IP
    'DOWNLOADER_MIDDLEWARES' : {
                'scrapy.downloadermiddlewares.useragent.UserAgentMiddleware': None,
                'scrapy.downloadermiddlewares.retry.RetryMiddleware': None,
                'scrapy_fake_useragent.middleware.RandomUserAgentMiddleware': 400,
                'scrapy_fake_useragent.middleware.RetryUserAgentMiddleware': 401,
            },
    'FAKEUSERAGENT_PROVIDERS' : [
                'scrapy_fake_useragent.providers.FakeUserAgentProvider',
                'scrapy_fake_useragent.providers.FakerProvider',  
                'scrapy_fake_useragent.providers.FixedUserAgentProvider',
            ],
    "REQUEST_FINGERPRINTER_IMPLEMENTATION": "2.7",
    "TWISTED_REACTOR": "twisted.internet.asyncioreactor.AsyncioSelectorReactor",
    "FEED_EXPORT_ENCODING": "utf-8",
    "CUSTOM_LOG_EXTENSION":True,
    "EXTENSIONS":{
        'scrapy.extensions.telnet.TelnetConsole': None,
        'ClappScrapers.den_rental.spider.extension.CustomLogExtension': 1,},
    'RETRY_ENABLED': True,
    'RETRY_TIMES': 5,  # Number of times to retry
    'RETRY_HTTP_CODES': [500, 502, 503, 504, 522, 524, 408, 429, 403, 407],
}


# Define your item pipelines here
#
# Don't forget to add your pipeline to the ITEM_PIPELINES setting
# See: https://docs.scrapy.org/en/latest/topics/item-pipeline.html


# useful for handling different item types with a single interface
from itemadapter import ItemAdapter
from datetime import datetime
import time
import json
import re
from scrapy.signalmanager import dispatcher
from scrapy import signals

def normalize_key (text):

    if text is None:
        return ''
    
    if text is not None:
        
        normalized_key = text.replace('(p.m.)','').replace('(s)','').strip()

        # Replace spaces with underscores
        normalized_key = normalized_key.replace(' ', '_').replace('.','').replace('-','_').strip()
    
        # Remove special characters
        normalized_key = ''.join(char for char in normalized_key if char.isalnum() or char in ['_', '-'])

        normalized_key = normalized_key.replace('__',"_")

        # Decapitalize the key
        normalized_key = normalized_key.lower()

        return normalized_key

class MergedDataPipeline:
    def __init__(self):
        self.raw_data = []
        self.cleaned_data = []
    def process_item(self,item,spider):
        adapter = ItemAdapter(item)

        if adapter.get('source'):
            #Process raw data
            self.raw_data.append(adapter.asdict())

            #Process cleaned data
            cleaned_item = self.clean_item(item)
            self.cleaned_data.append(cleaned_item)
        return item
    def close_spider(self,spider):

        #Process data
        spider.log("pipeline Closing spider and sending custom_closed_signal...")
        dispatcher.send(signal=signals.spider_closed, 
                        sender=spider, 
                        cleaned_data = self.cleaned_data, 
                        raw_data = self.raw_data)
    def clean_item (self,item):

        if isinstance(item, dict):
            cleaned_item = {}
            for key, value in item.items():
                cleaned_value = self.clean_value(key, value)
                if cleaned_value is not None:
                    cleaned_item[normalize_key(key)] = cleaned_value
            return cleaned_item
        elif isinstance(item,list):
            cleaned_list = []
            for element in item:
                cleaned_element = self.clean_item(element)
                if cleaned_element:
                    cleaned_list.append(cleaned_element)
            return cleaned_list
        else:
            return item
        
    def clean_value(self, key, value):
        
        date_keys = ['districtPlanFrom','district_plan_from']

        float_keys = ['property_living_area_sqm','property_monthly_expense','latitude','longitude','property_plot_area_sqm','property_sale_price_per_sqm','property_sale_price','property_weighted_area_sqm','property_latest_valuation',]

        int_keys = ['property_energy_label_year','property_floor_number','property_rooms_amount','property_toilets_amount','property_price_change_percentage','property_construction_year']

        str_keys = ['address_house_number']

        if key in date_keys:

            return self.convert_to_unix_timestamp(value)
            
        elif key in float_keys:

            return self.to_float(value)
        
        elif key in int_keys:

            return self.to_integer(value)
        
        elif key in str_keys:
            
            return self.to_string(value)
        else:

            return value
    
    def convert_to_unix_timestamp(self,date_str):
        try:
            # Specify the format for "01 February 2024"
            date_formats = ["%d %B %Y", "%d %b %Y", "%d/%m/%Y","%Y-%m-%d"]  # You can add more formats as needed
            for date_format in date_formats:
                try:
                    date_object = datetime.strptime(date_str, date_format)
                    unix_timestamp = int(date_object.timestamp())
                    return unix_timestamp
                except ValueError:
                    pass

            # If the input is not in the specified date formats, try parsing it as datetime string
            dt_formats = ["%Y-%m-%dT%H:%M:%S.%f%z", "%Y-%m-%dT%H:%M:%S%z"]  # Add more formats as needed
            for dt_format in dt_formats:
                try:
                    dt_obj = datetime.strptime(date_str, dt_format)
                    timestamp = round(dt_obj.timestamp())
                    return timestamp
                except ValueError:
                    pass

            return None  # Return None if no valid format is found
        except Exception as e:
            return None

        
    def to_boolean(self, value):
        if value == "Yes":
            return True
        elif value == "No":
            return False
        else:
            return None

    def to_integer(self, value):
        try:
            return int(value)
        except ValueError:
            return None
        
    def to_float(self, value):
        if value is None:
            return
        try:
            return float(value)
        except ValueError:
            return None
        
    def to_string(self, value):
        if value is None:
            return
        try:
            return str(value)
        except ValueError:
            return None
        
# class RawDataPipeline:

#     def __init__(self):
#         self.raw_data = []

#     def process_item(self, item, spider):
#         # Basic data validation: Check if the scraped item is not empty
#         adapter = ItemAdapter(item)
#         if adapter.get('source'):
#             self.raw_data.append(adapter.asdict())
#         return item

#     def close_spider(self, spider):
#         with open('boligsiden_raw_data.json', 'w',encoding='utf-8') as file:
#             json.dump(self.raw_data, file, indent=2, ensure_ascii=False)


# class CleanedDataPipeline:

#     def __init__(self):
#         self.cleaned_data = []

#     def process_item(self, item, spider):

#         cleaned_item = self.clean_item(item)
#         self.cleaned_data.append(cleaned_item)
#         return item

#     def close_spider(self, spider):

#         with open('boligsiden_cleaned_data.json', 'w',encoding='utf-8') as file:
#             json.dump(self.cleaned_data, file, indent=2, ensure_ascii=False)

#     def clean_item (self,item):

#         if isinstance(item, dict):
#             cleaned_item = {}
#             for key, value in item.items():
#                 cleaned_value = self.clean_value(key, value)
#                 if cleaned_value is not None:
#                     cleaned_item[normalize_key(key)] = cleaned_value
#             return cleaned_item
#         elif isinstance(item,list):
#             cleaned_list = []
#             for element in item:
#                 cleaned_element = self.clean_item(element)
#                 if cleaned_element:
#                     cleaned_list.append(cleaned_element)
#             return cleaned_list
#         else:
#             return item
        
#     def clean_value(self, key, value):
        
#         date_keys = ['districtPlanFrom','district_plan_from']

#         if isinstance(value,str):

#             if key in date_keys:

#                 return self.convert_to_unix_timestamp(value)
            
#             if key == 'property_energy_label':
#                 match = re.match(r'([A-Z])(\d{4})?', value)
#                 if match:
#                     label = match.group(1) if match.group(1) else None
#                     year = int(match.group(2)) if match.group(2) else None

#                 cleaned_value = {

#                 f"{key}": label,
#                 f"{key}_year": year,
#             }
#                 return cleaned_value
            
#             else :
#                 return value
                
#         else:

#             return value
    
#     def convert_to_unix_timestamp(self,date_str):
#         try:
#             # Specify the format for "01 February 2024"
#             date_formats = ["%d %B %Y", "%d %b %Y", "%d/%m/%Y","%Y-%m-%d"]  # You can add more formats as needed
#             for date_format in date_formats:
#                 try:
#                     date_object = datetime.strptime(date_str, date_format)
#                     unix_timestamp = int(date_object.timestamp())
#                     return unix_timestamp
#                 except ValueError:
#                     pass

#             # If the input is not in the specified date formats, try parsing it as datetime string
#             dt_formats = ["%Y-%m-%dT%H:%M:%S.%f%z", "%Y-%m-%dT%H:%M:%S%z"]  # Add more formats as needed
#             for dt_format in dt_formats:
#                 try:
#                     dt_obj = datetime.strptime(date_str, dt_format)
#                     timestamp = round(dt_obj.timestamp())
#                     return timestamp
#                 except ValueError:
#                     pass

#             return None  # Return None if no valid format is found
#         except Exception as e:
#             return None

        
#     def to_boolean(self, value):
#         if value == "Yes":
#             return True
#         elif value == "No":
#             return False
#         else:
#             return None

#     def to_integer(self, value):
#         try:
#             return int(value.replace(',', '').replace('.', ''))
#         except ValueError:
#             return None
        
#     def to_float(self, value):
#         if value is None:
#             return
#         try:
#             return float(value.replace('.','').replace(',','.').replace(' kr', '').replace(' m²', ''))
#         except ValueError:
#             return None
    




settings = {'BOT_NAME' : "denmarksell_scraper",
            'SPIDER_MODULES' : ["ClappScrapers.den_boligsiden_sell.spider"],
            'NEWSPIDER_MODULE' : "ClappScrapers.den_boligsiden_sell.spider",
            'LOG_LEVEL':'WARNING',
            'ROBOTSTXT_OBEY' : False,
            'COOKIES_ENABLED' : False,
            'CONCURRENT_REQUESTS' : 8,  # Reduce the number of concurrent requests
            'CONCURRENT_REQUESTS_PER_DOMAIN' : 4,  # Reduce the number of concurrent requests per domain
            'CONCURRENT_REQUESTS_PER_IP' : 4,  # Reduce the number of concurrent requests per IP
            'ITEM_PIPELINES' : {
                "ClappScrapers.den_boligsiden_sell.spider.pipelines.MergedDataPipeline": 200
            },
            'DOWNLOADER_MIDDLEWARES' : {
                'scrapy.downloadermiddlewares.useragent.UserAgentMiddleware': None,
                'scrapy.downloadermiddlewares.retry.RetryMiddleware': None,
                'scrapy_fake_useragent.middleware.RandomUserAgentMiddleware': 400,
                'scrapy_fake_useragent.middleware.RetryUserAgentMiddleware': 401,
            },

            'FAKEUSERAGENT_PROVIDERS' : [
                'scrapy_fake_useragent.providers.FakeUserAgentProvider',
                'scrapy_fake_useragent.providers.FakerProvider',  
                'scrapy_fake_useragent.providers.FixedUserAgentProvider',
            ],

            'REQUEST_FINGERPRINTER_IMPLEMENTATION' : "2.7",
            'TWISTED_REACTOR' : "twisted.internet.asyncioreactor.AsyncioSelectorReactor",
            'FEED_EXPORT_ENCODING' : "utf-8",
            "CUSTOM_LOG_EXTENSION":True,
            "EXTENSIONS":{
                'scrapy.extensions.telnet.TelnetConsole': None,
                'ClappScrapers.asunnot.spider.extension.CustomLogExtension': 1,},
            'RETRY_ENABLED': True,
            'RETRY_TIMES': 5,  # Number of times to retry
            'RETRY_HTTP_CODES': [500, 502, 503, 504, 522, 524, 408, 429, 403, 407],
}


# Define your item pipelines here
#
# Don't forget to add your pipeline to the ITEM_PIPELINES setting
# See: https://docs.scrapy.org/en/latest/topics/item-pipeline.html


# useful for handling different item types with a single interface
from itemadapter import ItemAdapter
from datetime import datetime
import time
import json
import re
from scrapy.signalmanager import dispatcher
from scrapy import signals

class MergedDataPipeline:
    def __init__(self):
        self.raw_data = []
        self.cleaned_data = []
    def process_item(self,item,spider):
        adapter = ItemAdapter(item)

        if adapter.get('source'):
            #Process raw data
            self.raw_data.append(adapter.asdict())

            #Process cleaned data
            cleaned_item = self.clean_item(item)
            self.cleaned_data.append(cleaned_item)
        return item
    def close_spider(self,spider):

        #Process data
        spider.log("pipeline Closing spider and sending custom_closed_signal...")
        dispatcher.send(signal=signals.spider_closed, 
                        sender=spider, 
                        cleaned_data = self.cleaned_data, 
                        raw_data = self.raw_data)
    def clean_item (self,item):

        if isinstance(item, dict):
            cleaned_item = {}
            for key, value in item.items():
                cleaned_value = self.clean_value(key, value)
                if cleaned_value is not None:
                    cleaned_item[key] = cleaned_value
            return cleaned_item
        elif isinstance(item,list):
            cleaned_list = []
            for element in item:
                cleaned_element = self.clean_item(element)
                if cleaned_element:
                    cleaned_list.append(cleaned_element)
            return cleaned_list
        else:
            return item
        
    def clean_value(self, key, value):
        
        date_keys = ['timeline_advertisement_creation']


        if isinstance(value,str):

            if key in date_keys:

                return self.convert_to_unix_timestamp(value)
            
            else :
                return value
                
        else:

            return value
    
    def convert_to_unix_timestamp(self,date_str):
        try:
            # Specify the format for "01 February 2024"
            date_formats = ["%d %B %Y", "%d %b %Y", "%d/%m/%Y","%Y-%m-%d"]  # You can add more formats as needed
            for date_format in date_formats:
                try:
                    date_object = datetime.strptime(date_str, date_format)
                    unix_timestamp = int(date_object.timestamp())
                    return unix_timestamp
                except ValueError:
                    pass

            # If the input is not in the specified date formats, try parsing it as datetime string
            dt_formats = ["%Y-%m-%dT%H:%M:%S.%f%z", "%Y-%m-%dT%H:%M:%S%z"]  # Add more formats as needed
            for dt_format in dt_formats:
                try:
                    dt_obj = datetime.strptime(date_str, dt_format)
                    timestamp = round(dt_obj.timestamp())
                    return timestamp
                except ValueError:
                    pass

            return None  # Return None if no valid format is found
        except Exception as e:
            return None

        
    def to_boolean(self, value):
        if value == "Yes":
            return True
        elif value == "No":
            return False
        else:
            return None

    def to_integer(self, value):
        try:
            return int(value.replace(',', '').replace('.', ''))
        except ValueError:
            return None
        
    def to_float(self, value):
        if value is None:
            return
        try:
            return float(value.replace('.','').replace(',','.').replace(' kr', '').replace(' m²', ''))
        except ValueError:
            return None

    def map_floor_to_integer(self, floor):
        floor = floor.strip().lower()
        if floor == "ground floor":
            return 0
        elif floor[:-2].isdigit():
            return int(floor[:-2])
        else:
            return None


from urllib.parse import urlencode
from scrapy import Request
class ScrapeOpsProxyMiddleware:

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler.settings)


    def __init__(self, settings):
        self.scrapeops_api_key = settings.get('SCRAPEOPS_API_KEY')
        self.scrapeops_endpoint = 'https://proxy.scrapeops.io/v1/?'
        self.scrapeops_proxy_active = settings.get('SCRAPEOPS_PROXY_ENABLED', False)


    @staticmethod
    def _param_is_true(request, key):
        if request.meta.get(key) or request.meta.get(key, 'false').lower() == 'true':
            return True
        return False


    @staticmethod
    def _replace_response_url(response):
        real_url = response.headers.get(
            'Sops-Final-Url', def_val=response.url)
        return response.replace(
            url=real_url.decode(response.headers.encoding))
    

    def _get_scrapeops_url(self, request):
        payload = {'api_key': self.scrapeops_api_key, 'url': request.url}
        if self._param_is_true(request, 'sops_render_js'):
            payload['render_js'] = True
        if self._param_is_true(request, 'sops_residential'): 
            payload['residential'] = True
        if self._param_is_true(request, 'sops_keep_headers'): 
            payload['keep_headers'] = True
        if request.meta.get('sops_country') is not None:
            payload['country'] = request.meta.get('sops_country')
        proxy_url = self.scrapeops_endpoint + urlencode(payload)
        return proxy_url


    def _scrapeops_proxy_enabled(self):
        if self.scrapeops_api_key is None or self.scrapeops_api_key == '' or self.scrapeops_proxy_active == False:
            return False
        return True
    
    def process_request(self, request, spider):
        if self._scrapeops_proxy_enabled is False or self.scrapeops_endpoint in request.url:
            return None
        
        scrapeops_url = self._get_scrapeops_url(request)
        new_request = request.replace(
            cls=Request, url=scrapeops_url, meta=request.meta)
        return new_request


    def process_response(self, request, response, spider):
        new_response = self._replace_response_url(response)
        return new_response
